// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

/**
 * The Caesar algorithm is supposedly the one Julius Caesar used by
 * hand many centuries ago. As you can imagine, this is <b>NOT A STRONG
 * CIPHER</b>, but included only to show how to write a very simple
 * Key class for the Cryptonite package. Often, the first assignment given
 * to cryptography students is to break this cipher.
 * <p>
 * Data is encrypted byte-by-byte by adding a constant value to it and
 * taking the 8 lowest order bits.
 * <p>
 * The CDS for a CaesarKey object is <code>CaesarKey(n)</code> where
 * n is a value in the range 0..255.
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public final class CaesarKey extends SymmetricKey implements CipherKey {

    private byte shift;
    
    /** Create a new Caesar key with the specified shift. */
    public CaesarKey(byte shift) {
        this.shift = shift;
    }

    /** Create a new random Caesar key. */
    public CaesarKey() {
        shift = (byte)random.nextInt();;
    }

    /**
     * Called by <code>Cryptonite.fromString(String)</code> .
     *
     * @exception InvalidCDSException if the CDS is in some way malformed.
     */
    public CaesarKey(String shift) throws InvalidCDSException {
        int i=0;
        while((i<shift.length()) && Character.isWhitespace(shift.charAt(i)))
            i++;
        if(i==shift.length())
            throw new InvalidCDSException("Empty argument passed to CaesarKey(String)");
        int j=shift.length()-1;
        while((j>=0) && Character.isWhitespace(shift.charAt(j)))
            j--;
        // shift[i..j] is the contents of shift[] with spaces stripped.
        if(j==i && shift.charAt(i)=='?')
            this.shift = (byte)random.nextInt();
        else
            this.shift = (byte)Integer.parseInt(shift.substring(i,j+1));
    }

    /** Return the size of the key in bits. */
    public int getSize(){
        return 8;
    }

    /** The block-size for the Caesar cipher is one byte. */
    public final int plainBlockSize(){
	return 1;
    }
    
    /** The block-size for the Caesar cipher is one byte. */
    public final int cipherBlockSize(){
        return 1;
    }
    
    /** Return the name of the algorithm used by this key. */
    public String getAlgorithm(){
        return ("Caesar");
    }

    /** Return true if the two keys are eqivalent. */
    public final boolean equals(Object o){
        if (o.getClass() != this.getClass())
            return false;
        return shift==((CaesarKey)o).shift;
    }

    /** Return a CDS for this key. */
    public String toString(){
        if(shift<0)
            return "CaesarKey("+(256+shift)+")";
        else
            return "CaesarKey("+shift+")";
    }

    /** Encrypt <code>source[i]</code> and put it <code>dest[j]</code>. */
    public final void encrypt(byte[] source, int i, byte[] dest, int j){
        dest[j]=(byte)(source[i]+shift);
    }

    /** Decrypt <code>source[i]</code> and put it <code>dest[j]</code>. */
    public final void decrypt(byte[] source, int i, byte[] dest, int j){
	dest[j]=(byte)(source[i]-shift);
    }

}
