// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.math.BigInteger;
import java.io.*;

/**
 * This object holds one Diffie-Hellman key. It is not used for encryption
 * or signatures, but only to initialize Difie-Hellman key-exchange objects.
 * <p>
 * CDS for a Diffie-Hellman key is <code>DHKey(a,g,m,pub)</code>
 * for a public key or <code>DHKey(a,g,m,pri)</code> for a private
 * key. In both cases <code>a</code>,<code>g</code> and <code>m</code> are
 * hexadecimal numbers.
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class DHKey extends K implements Key {
    
    /**
     * Actual key value. */
    private BigInteger key;
    
    /**
     * Generator */
    private BigInteger g;
    
    /**
     * Modulus */
    private BigInteger m;
    
    /**
     * Private-key flag. */
    private boolean pri;
    
    /**
     * If <code>pri</code>, then this is the public key
     * or <code>null</code>. */
    private BigInteger pubKey;
    
    /**
     * Create new Diffie-Hellman key for <code>a</code> in the group
     * modulo <code>m</code> with generator <code>g</code>. It is a
     * private key iff </code>pri<code> is <code>true</code>. */
    public DHKey(BigInteger a, BigInteger g, BigInteger m, boolean pri){
        key=a;
        this.g=g;
        this.m=m;
        this.pri=pri;
    }
    
    /**
     * Called by <code>Cryptonite.fromString(String)</code> .
     *
     * @exception InvalidCDSException if the CDS is in some way malformed.
     */
    protected DHKey(String arg) throws InvalidCDSException{
        try {
            StreamTokenizer st=new StreamTokenizer(new StringReader(arg));
            st.ordinaryChars('0','9');
            st.wordChars('0','9');
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as first argument to DHKey()");
            key = new BigInteger(st.sval,16);
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+key.toString(16));
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as second argument to DHKey()");
            g = new BigInteger(st.sval,16);
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+g.toString(16));
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as third argument to DHKey()");
            m = new BigInteger(st.sval,16);
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+m.toString(16));
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("\"pub\" or \"pri\" expected as second argument to DHKey()");
            String pubpri=st.sval;
            if(pubpri.equals("pub"))
                pri=false;
            else if(pubpri.equals("pri"))
                pri=true;
            else
                throw new InvalidCDSException("\"pub\" or \"pri\" expected as second argument to DHKey()");
            if(st.nextToken()!=StreamTokenizer.TT_EOF){
                System.out.println(st.ttype);
                throw new InvalidCDSException("no more parameters expected in DHKey() after "+pubpri);
            }
        } catch (IOException e){
            // StreamReader doesn't really throw exceptions.
        }
    }

    /**
     * Return the largest prime <code>p</code> &lt; 2^<code>n</code>
     * such that (<code>p</code>-1)/2 is prime. */
    public static BigInteger getModulus(int n){
        if(n<256)
            n=256;
        else
            n=((n+7)/8)*8; // round up to nearest byte.
        
        if (n==256) 
	    return TWO.pow(n).subtract(BigInteger.valueOf(36113));
	if (n==512)
            return TWO.pow(n).subtract(BigInteger.valueOf(38117));
        if (n==1024)
            return TWO.pow(n).subtract(BigInteger.valueOf(1093339));
        if (n==2048)
            return TWO.pow(n).subtract(BigInteger.valueOf(1942289));

        BigInteger m=TWO.pow(n).subtract(ONE);
        int sub=1;
        BigInteger k=m.divide(TWO);
        while(!m.isProbablePrime(2) || // First check very quickly
              !k.isProbablePrime(2) ||
              !m.isProbablePrime(4) ||
              !k.isProbablePrime(4) ||
              !m.isProbablePrime(8) ||
              !k.isProbablePrime(8) ||
              !m.isProbablePrime(primeCertainty-14)|| // Then check very well
              !k.isProbablePrime(primeCertainty-14)){
            m=m.subtract(FOUR);
            sub+=4;
            k=m.divide(TWO);
        }
        return m;
    }

    /**
     * Returns the smallest number <code>g</code> > 1 with
     * gcd(<code>g</code>,<code>m</code>)=1. */
    public static BigInteger getGenerator(BigInteger m){
        BigInteger g=TWO;
        while(!m.gcd(g).equals(ONE))
            g=g.add(ONE);
        return g;
    }

    /**
     * Create a random private DHKey with an <code>n</code
     * bit modulus.
     * <p>

     * Pe-calcualted modula exist for bit lengths 256, 512 and 1024. (The 2048 bit
     * modulus is still being calculated. Using these values saves a
     * lot of time and does not weaken the keys. */
    public DHKey(int n){
        n = (n<256) ? 256 : (((n+7)/8)*8); // round up to nearest byte.

        m=getModulus(n);
        g=getGenerator(m);
        do{
            key=new BigInteger(n,random);
        } while (key.compareTo(m)>0);

        pri=true;
    }
    
    /**
     * Create a pair of public/private keys in a group with an
     * <code>n</code> bit modulo.
     * <p>
     * Pe-calcualted modula exist for bit lengths 256, 512, 1024 and
     * 2048. Using these values saves a lot of time and does not weaken
     * the keys. */
    public static KeyPair createKeys(int n){
        DHKey pri = new DHKey(n);
        return new KeyPair(pri.getPublic(),
                           pri);
    }

    /**
     * Return the "size" of the key. This is a measure of how 
     * difficult it is to break and is heavily dependant on
     * the algorithm used. In many cases it will be the
     * actual number of bits needed to store the key. */
    public int getSize(){
        return m.bitLength();
    }
    
    /** Return the name of the algorithm used by this key. */
    public String getAlgorithm(){
        return "Diffie-Hellman";
    }
    
    /**
     * Returns true iff this is a private key. The default behaviour is to
     * always return true. Classes for asymmetric keys should override
     * this. */
    public boolean isPrivate(){
        return pri;
    }
    
    /**
     * Return the BigInteger representing the public key from the pair. */
    public DHKey getPublic(){
        if(!pri)
            return this;
        if(pubKey==null)
            pubKey = g.modPow(key,m);
        return new DHKey(pubKey,g,m,false);
    }
    
    /**
     * Return the BigInteger representing this key. */
    public BigInteger getKey(){
        return key;
    }

    /**
     * Return the modulus for this key. */
    public BigInteger getM(){
        return m;
    }
    
    /**
     * Return the generator for this key. */
    public BigInteger getG(){
        return g;
    }
    
    /** 
     * Calculate the fingerprint for this key or the other in the pair. */
    protected Fingerprint calcFingerprint(boolean other){
        // The key-pair is uniquely defined by the public key, since the
        // private key is "merely" the dicrete log og the public key.
        FingerprintState fs=new MD5State();
        fs.update(getPublic().getKey().toByteArray());
        if(other==pri)
            fs.update("pub");
        else
            fs.update("pri");
        return fs.calculate();
    }
    
    /**
     * Return true if the two keys are eqivalent. */
    public final boolean equals(Object o){
        if (o.getClass() != this.getClass())
            return false;
        DHKey key=(DHKey)o;
        return key.toString()==toString();
    }
    
    /**
     * Returns true if this and key are a matched pair
     * of public/private keys or the same symmetric key.
     * The default behaviour is to return the same value
     * as equals(key). */
    public boolean matches(Key key){
        return equals(key);
    }
    
    /**
     * Return a CDS for this key. */
    public String toString(){
        StringBuffer sb=new StringBuffer();
        sb.append("DHKey(");
        sb.append(key.toString(16));
        sb.append(',');
        sb.append(g.toString(16));
        sb.append(',');
        sb.append(m.toString(16));
        sb.append(',');
        sb.append(pri ? ",pri)" : ",pub)");
        return sb.toString();
    }
    
}
