// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.io.*;

/**
 * This OutputStream encrypts everything written to it using the key suggested
 * by the key-exchange object passed to the constructor and in the mode passed
 * to the constructor.
 * <p>
 * Beware that up to one byte less than a whole block of garbage may be
 * appended to the data when flush is called, such as when the stream is closed.
 *
 * @see is.hi.logir.cryptonite.DecryptStream
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class EncryptStream extends FilterOutputStream{
  
  private EncryptMode emode;
  
  /**
   * Create a new EncryptStream which writes encrypted data to
   * <code>out</code>.  The <code>emode</code> object will be used
   * for encryption and if <code>kex</code> is not <code>null</code>
   * it will be used to inform the appropriate reader about what key
   * to use for decryption.
   * <p>
   * if <code>kex</code> is not <code>null</code>, then the key
   * returned by its <code>sessionKey()</code> method will be used in
   * the <code>emode</code> object. Otherwise, <code>emode</code> is
   * assumed to contain the proper key.
   * <p> 
   * The data output can be read by a DecryptStream object which has
   * been initialized with the DecryptMode object which corresponds to
   * <code>emode</code> and the NoninteractiveKeyExchangeServer object
   * which corresponds <code>kex</code>, or <code>null</code> if
   * <code>kex</code> is <code>null</code>.
   *
   * @exception ProtocolException if a problem arises with the key-exchange protocol.
   * @exception IOException if problems arise with the underlying OutputStream.  */
  public EncryptStream(OutputStream out, NoninterKeyExClient kex, EncryptMode emode) throws ProtocolException, IOException{
    super(out);
    this.emode = emode;
    
    if(kex!=null){
      // Let's do key-exchange
      byte[] msg = kex.message(null);
      int l = msg.length;
      out.write(l >>> 24);
      out.write(l >>> 16);
      out.write(l >>>  8);
      out.write(l       );
      out.write(msg);
      
      try{
	CipherKey key=(CipherKey)kex.sessionKey();
	if(key==null)
	  throw new ProtocolException("A non-interactive protocol should only need one message.");
	emode.setKey(key);
      } catch (ClassCastException e){
	throw new ProtocolException("The key-exchange protocol proposes to use a non-cipher key for encryption.");
      }
    }
    
  }
  
  /** Writes the specified byte to this output stream. */
  public synchronized void write(int b) throws IOException{
    byte[] buf = { (byte)b };
    write(buf,0,1);
  }
  
  /**
   * Writes len bytes from the specified byte array starting at offset
   * off to this output stream. 
   */
  public synchronized void write(byte[] buf, int off, int len) throws IOException{
    byte[] ctext = emode.encrypt(buf,off,len);
    out.write(ctext,0,ctext.length);
  }
  
  /**
   * Flushes this output stream and forces any buffered output bytes to
   * be written out to the stream. If the number of bytes written is not
   * a multiple of the plainBlockSize of the Key used for encryption, up
   * to one byte less than a whole block of garbage may be appended to
   * the data when flush is called.
   */
  public synchronized void flush() throws IOException{
    byte[] ctext = emode.flush();
    out.write(ctext,0,ctext.length);
    out.flush();
  }
  
}
