// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.io.*;
/**
 * This class is used to hold a fingerprint of a particular data buffer.
 * The idea is to calculate a fingerprint in such a way that it is extremely
 * difficult to create another buffer that gives the same fingerprint. If
 * that buffer also has to match some other criteria, such as being a valid 
 * text file in a particular language, then it becomes next to impossible.
 * <p>
 * All this depends on the hash function used to create the fingerprint
 * being a good one. Fingerprints are created by the various subclasses
 * of FingerprintState, so you should look there for information about
 * a particular hash function.
 * <p>
 * You should be very careful about saving fingerprints. If a fingerprint
 * is not calculated at run-time, then it should not be trusted, since it
 * may have been tampered with in storage or transit.
 * <p>
 * The CDS for a Fingerprint object is <code>Fingerprint(NAME,FP)</code>
 * where <code>NAME</code> is the name of the algorithm used and
 * <code>FP</code> the actual fingerprint.
 *
 * @see is.hi.logir.cryptonite.FingerprintState
 * @see is.hi.logir.cryptonite.Signature
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class Fingerprint extends Cryptonite{

    // CREATE FINGERPRINTS DIRECTLY
    
    /**
     * Create a Fingerprint object for the data in
     * buf[offset..offset+length-1] with the named
     * fingerprint algorithm.
     *
     * @exception InvalidCDSException if a FingerprintState object for the
     *            named algorithm could not be created.
     */
    public static Fingerprint create(byte[] buf, int offset, int length, String algorithm) throws InvalidCDSException {
	FingerprintState fps = FingerprintState.create(algorithm);
	fps.update(buf, offset, length);
	return fps.calculate();
    }
    
    /**
     * Create a Fingerprint object for the data in buf
     * with the named fingerprint algorithm.
     *
     * @exception InvalidCDSException if a FingerprintState object for the
     *            named algorithm could not be created.
     */
    public static Fingerprint create(byte[] buf, String algorithm) throws InvalidCDSException {
	return create(buf, 0, buf.length, algorithm);
    }
    
    /**
     * Create a Fingerprint object for the string s
     * with the named fingerprint algorithm.
     *
     * @exception InvalidCDSException if a FingerprintState object for the
     *            named algorithm could not be created.
     */
    public static Fingerprint create(String s, String algorithm) throws InvalidCDSException {
	FingerprintState fps = FingerprintState.create(algorithm);
	fps.update(s);
	return fps.calculate();
    }

    // =================================================================== //
    
    /** Holds the actual bytes of the fingerprint value. */
    protected byte[]fp=null;

    /** Holds the name of the algorithm used to create this fingerprint. */
    protected String algorithm=null;

    /** 
     * Creates a new Fingerprint object with the given algorithm name and
     * the hash value from <code>fp[0..n-1]</code>. */
    public Fingerprint(String algorithm, byte[] fp, int n){
        this.algorithm = algorithm;
        this.fp = new byte[n];
        System.arraycopy(fp,0, this.fp,0, n);
    }
    
    /** 
     * Creates a new Fingerprint object with the given algorithm name and
     * the hash value from <code>fp</code>. */
    public Fingerprint(String algorithm, byte[] fp){
        this.algorithm = algorithm;
        this.fp = new byte[fp.length];
        System.arraycopy(fp,0, this.fp,0, fp.length);
    }

    /**
     * Called by <code>Cryptonite.fromString(String)</code> .
     *
     * @exception InvalidCDSException if a FingerprintState object for the
     *            named algorithm could not be created.
     */
    public Fingerprint(String arg) throws InvalidCDSException{
        try{
            StreamTokenizer st=new StreamTokenizer(new StringReader(arg));
            st.ordinaryChars('0','9');
            st.wordChars('0','9');
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Fingerprint algorithm name expected as first argument to Fingerprint()");
            algorithm=st.sval;
            if(st.nextToken()!=',')
                throw new InvalidCDSException(", expected after "+algorithm);
            if(st.nextToken()!=StreamTokenizer.TT_WORD)
                throw new InvalidCDSException("Hexadecimal string expected as second argument to Fingerprint()");
            fp=fromHexString(st.sval);
        } catch (IOException e){
            // StringReader doesn't actually throw any exceptions
        }
    }
    
    /** Return the name of the algorithm used for this fingerprint. */
    public String getName(){
        return algorithm;
    }

    /** Return an array of the bytes in the fingerprint. */
    public byte[] getBytes(){
        byte[] fp=new byte[this.fp.length];
        System.arraycopy(this.fp,0, fp,0, this.fp.length);
        return fp;
    }

    /** return true if <code>obj</code> is a fingerprint equal to <code>this</code>. */
    public boolean equals(Object obj){
        if (obj.getClass()!=this.getClass())
            return false;
        // It's the same class
        
        Fingerprint f=(Fingerprint)obj;
        if (!f.algorithm.equals(algorithm))
            return false;
        // It's the same algorithm.
        
        for (int i=fp.length-1; i>=0; i--)
            // this.fp[j] == fp[j] for j=0, 1, ..., i-1.
            if (fp[i]!=f.fp[i])
                return false;

        return true;
    }

    /**
     * Return a hash-code based on the bytes of the
     * fingerprint and the algorithm name.
     */
    public int hashCode(){
        int h=algorithm.hashCode();
        for (int i=0; i<fp.length; i++)
            h = h ^ ( fp[i] << (i%4)*8 );
        return h;
    }

    /** Return a CDS for this fingerprint. */
    public String toString(){
        return "Fingerprint("+algorithm+","+hexString(fp)+")";
    }
}
