// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.lang.reflect.*;

/**
 * An object of this class holds the status of a fingerprint still being
 * calculated.
 * <p>
 * A fingerprint state object can be repeatedly updated with data. At
 * any time a Fingerprint object can be requested for the data that has
 * then been added to the fingerprint state.
 * 
 * @see is.hi.logir.cryptonite.MD5State
 * @see is.hi.logir.cryptonite.Fingerprint
 * @see is.hi.logir.cryptonite.Signature
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public abstract class FingerprintState extends Cryptonite{
  
  // FINGERPRINT CLASS LIBRARIAN
  
  /**
   * Create a FingerprintState object for the named fingerprint
   * algorithm.
   *
   * @exception InvalidCDSException if a FingerprintState object for the
   *            named algorithm could not be created.
   */
  public static FingerprintState create(String algorithm) throws InvalidCDSException {
    // Construct an object
    Class cl;
    Constructor con;
    Object r;
    try{
      cl = Class.forName("is.hi.logir.cryptonite."+algorithm+"State");
    } catch (ClassNotFoundException e) {
      throw new InvalidCDSException("The class "+algorithm+"State was not found");
    }
    try {
      Class[] parType = new Class[0];
      con=cl.getConstructor(parType);
    } catch (Exception e) {
      throw new InvalidCDSException(algorithm+" does not have a "+algorithm+"State() constructor");
    }
    try{
      Object[] arg = new Class[0];
      r = con.newInstance(arg);
    } catch (InvocationTargetException e1){
      throw new InvalidCDSException("Unable to create an instance of "+algorithm+"State [ "+e1.getTargetException().toString()+" ]");
    } catch (Exception e){
      throw new InvalidCDSException("Unable to create an instance of "+algorithm);
    }
    try{
      return (FingerprintState) r;
    } catch (ClassCastException e){
      throw new InvalidCDSException(algorithm+"State is not a descendant of FingerprintState");
    }
  }    
  
  // INSTANCE METHODS

  /** Return the name of the algorithm used by this FingerprintState object. */
  public abstract String getName();
  
  /** Reset the state. */
  public abstract void reset();
  
  /**
   * Update the fingerprint state with the bytes from
   * <code>buf[offset, offset+length-1]</code>. */
  public abstract void update(byte[] buf, int offset, int length);
  
  /** Update the fingerprint state with the bytes from <code>buf</code>. */
  public void update(byte[] buf){
    update(buf, 0, buf.length);
  }
  
  /** Update the fingerprint state with the characters from <code>s</code>. */
  public void update(String s){
    update(s.getBytes());
  }
  
  /**
   * Return a fingerprint for the curret state, without
   * destroying the state. */
  public abstract Fingerprint calculate();

  /**
   * Return the number of bytes needed to make a valid hash. If a multiple
   * of this number of bytes is hashed, no padding is needed. If no such
   * value exists, returns 0. */
  public abstract int blockSize();

  /**
   * Returns the size of a fingerprint in bytes. */
  public abstract int hashSize();
}
