// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.util.*;
import java.io.*;

/**
 * This implementation of the KeySource interface stores keys in a
 * hash-table. It can be converted to and from a CDS for storage.
 * <p>
 * A database key-source would be more appropriate for large
 * collections of keys, but hasn't been written yet.
 * <p>
 * An interface to a key-server would be more appropriate for
 * really huge key collections but will have to wait even longer.
 * <p>
 * The CDS for a KeyRing object is <code>KeyRing(k1,k2,...,kn)</code>
 * with <code>n</code>>=0 and each <code>ki</code> the CDS for a
 * KeyRecord or Key object.
 *
 * @see is.hi.logir.cryptonite.Key
 * @see is.hi.logir.cryptonite.KeyRecord
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class KeyRing extends Cryptonite implements KeySource{

    Hashtable keys;
    
    /** Create empty key ring. */
    public KeyRing(){
        keys = new Hashtable();
    }

    /**
     * Called by <code>Cryptonite.fromString(String)</code> .
     *
     * @exception InvalidCDSException if the CDS is in some way malformed.
     */
    public KeyRing(String arg) throws InvalidCDSException{
        keys = new Hashtable();

        try{
            StringReader in=new StringReader(arg);
            while(true){
                Object k=fromString(in);
                if(k instanceof Key)
                    insert((Key)k);
                else if(k instanceof KeyRecord)
                    insert((KeyRecord)k);
                else
                    throw new InvalidCDSException("Only descendants of Key or KeyRecord expected");
                int ch=in.read();
                while(Character.isWhitespace((char)ch))
                    ch=in.read();
                if(ch==-1)
                    break;
                if(ch!=',')
                    throw new InvalidCDSException(", expected after "+k);
            }
        } catch (IOException e){
            // StringReader doesn't actually throw an IOException.
        }
    }

    /** Return a CDS for this KeyRing. */
    public String toString(){
        StringBuffer sb = new StringBuffer();
        sb.append("KeyRing(");
        
        Enumeration e=keys.elements();
        if(e.hasMoreElements())
            sb.append(e.nextElement().toString());
        while(e.hasMoreElements()){
            sb.append(",");
            sb.append(e.nextElement().toString());
        }
        sb.append(")");
              
        return sb.toString();
    }

    /**
     * Insert the Key <code>k</code> into the KeyRing. It will be wrapped in
     * a KeyRecord containing empty strings and no certificates.
     */
    public void insert(Key k){
        insert(new KeyRecord(k,"","",""));
    }

    /** Insert the KeyRecord <code>k</code> into the KeyRing. */
    public void insert(KeyRecord kc){
        keys.put(kc.getKey().getFingerprint(),kc);
    }

    /**
     * Return the key with the given fingerprint from the KeyRing if it is
     * found. Otherwise null is returned. <code>fingerprint</code> must be
     * created with the same algorithm as the the Key object uses, which
     * will be SHA1 for the Key classes included with Cryptonite.
     */
    public KeyRecord byFingerprint(Fingerprint fingerprint){
        return (KeyRecord)keys.get(fingerprint);
    }

}
