// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.math.BigInteger;

import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.StreamTokenizer;
import java.io.StringReader;

/**
 * The RSA algorithm is probably the best known and most widely used
 * public key algorithm. Breaking one RSA key is believed to be as
 * difficult as factoring the large integer that comprises the key, and
 * there is no known way to do this in a reasonable time. Therefore RSA
 * should be about as secure as anything if you keep your keys long. 1024
 * bits should be more than enough in most cases, but the truly paranoid may
 * want to use up to 4096 bit keys.
 * <p>
 * Each RSA key is a pair (r,n) of integers and matches another key (s,n).
 * If P is a block of plain data represented as an integer smaller than n,
 * then it can be encrypted with the transformation:
 * <blockquote>
 *   <code>E = (P^r) mod n</code>
 * </blockquote>
 * which has the inverse transformation:
 * <blockquote>
 *   <code>P = (E^s) mod n</code>
 * </blockquote>
 * <p>
 * The key owner will keep one key secret and publish the other as widely
 * as possible. This allows anyone who gets hold of the public key to
 * encrypt data which can only be decrypted with the corresponding private
 * key.
 * <p>
 * Data that is encrypted with a private key can similarly only be
 * decrypted with the corresponding public key. This is useful for digital
 * signatures.
 * <p>
 * When P is created from an array of bytes, it will correspond to as many
 * bytes of plain data as the bytes needed to store n, less one.
 * <p>
 * Each chunk of data encrypted with RSAKey has as many bytes as the key modulo.
 * however, the plain data it encodes has one less byte.
 * <p>
 * The CDS for the RSAKey class is <code>RSAKey(r,n,pub)</code> for a public key
 * or <code>RSAKey(r,n,pri)</code> for a private key. In both cases <code>r</code>
 * and <code>n</code> are hexadecimal numbers.
 *
 * @see is.hi.logir.cryptonite.Signature
 * @see is.hi.logir.cryptonite.Cryptonite#fromString
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class RSAKey extends K implements CipherKey,SignatureKey {
  
  /** <code>R</code> is the exponent used in all created public keys. */
  private static final BigInteger R=BigInteger.valueOf(65537);
  
  /** The RSA key exponent */
  private BigInteger r;
  
  /** The RSA key modulo */
  private BigInteger n;
  
  /** Is it a private key? */
  private boolean pri;
  
  /** The fingerprint for the other key in the pair, or null. */
  private Fingerprint matchPrint=null;
  
  ///////////////////////////////////////////////////////////////////////
  // KEY MANAGEMENT CODE
  
  /**
   * Create a new RSA key <code>(r,n)</code>.
   * It is a private key if <code>pri</code> is true.
   */
  public RSAKey(BigInteger r, BigInteger n, boolean pri) {
    super();
    this.pri = pri;
    this.r = r;
    this.n = n;
  }
  
  /**
   * Called by <code>Cryptonite.fromString(String)</code> .
   *
   * @see is.hi.logir.cryptonite.Cryptonite#fromString
   *
   * @exception InvalidCDSException if the CDS is in some way malformed.
   */
  public RSAKey(String arg) throws InvalidCDSException{
    try {
      StreamTokenizer st=new StreamTokenizer(new StringReader(arg));
      st.ordinaryChars('0','9');
      st.wordChars('0','9');
      if(st.nextToken()!=StreamTokenizer.TT_WORD)
        throw new InvalidCDSException("Hexadecimal string expected as first argument to RSAKey()");
      r = new BigInteger(st.sval,16);
      if(st.nextToken()!=',')
        throw new InvalidCDSException(", expected after "+r.toString(16));
      if(st.nextToken()!=StreamTokenizer.TT_WORD)
        throw new InvalidCDSException("Hexadecimal string expected as second argument to RSAKey()");
      n = new BigInteger(st.sval,16);
      if(st.nextToken()!=',')
        throw new InvalidCDSException(", expected after "+n.toString(16));
      if(st.nextToken()!=StreamTokenizer.TT_WORD)
        throw new InvalidCDSException("\"pub\" or \"pri\" expected as third argument to RSAKey()");
      String pubpri=st.sval;
      if(pubpri.equals("pub"))
        pri=false;
      else if(pubpri.equals("pri"))
        pri=true;
      else
        throw new InvalidCDSException("\"pub\" or \"pri\" expected as third argument to RSAKey()");
      if(st.nextToken()!=StreamTokenizer.TT_EOF){
        System.out.println(st.ttype);
        throw new InvalidCDSException("no more parameters expected in RSAKey after "+pubpri);
      }
    } catch (IOException e){
      // StreamReader doesn't really throw exceptions.
    }
  }
  
  /**
   * Create a pair of public/private keys. The key modulo will be
   * <code>bitLength</code> or <code>bitLength-1</code> bits.
   */
  public static KeyPair createKeys(int bitLength){
    if (bitLength<16)
      bitLength=16;
    
    BigInteger s=null;
    BigInteger n=null;
    while(n==null) {
      // FIXME! Pick random odd number and check every second
      //        value from there, as per Brandt & Damgaard
      // FIXME! Use java.util.Random to test for primality.
      //        (but of course use good random bits to select
      //         the starting point)
      try{
        BigInteger p=new BigInteger(bitLength/2, primeCertainty, random);
        BigInteger q=new BigInteger(bitLength/2, primeCertainty, random);
        BigInteger m=p.subtract(ONE).multiply(q.subtract(ONE));
        s=R.modInverse(m);
        n=p.multiply(q);
      } catch (ArithmeticException e) {
        // Key generation failed... just try again.
        n=null;
      }
    }
    
    Key pub=new RSAKey(R,n, false);
    Key pri=new RSAKey(s,n, true);
    return new KeyPair(pub,pri);
  }
  
  /**
   * Create a KeyPair object holding objects for the public RSA key
   * <code>(r,n)</code> and the private RSA key (s,n).
   *
   * @exception KeyException if (r,s,n) does not describe a valid RSA key.
   */
  public static KeyPair createKeys(BigInteger r, BigInteger s, BigInteger n) throws KeyException {
    Key pub = new RSAKey(r,n, false);
    Key pri = new RSAKey(s,n, true);
    try{
      return new KeyPair(pub,pri);
    } catch (Exception e) {
      throw new KeyException("The (r,s,n) triplet does not specify a matching pair of RSA keys.");
    }
  }
  
  /** Return the size of the key modulo in bits. */
  public int getSize(){
    return n.bitLength();
  }
  
  /** Return "RSA" */
  public String getAlgorithm(){
    return ("RSA");
  }
  
  /** 
   * Calculate the fingerprint for this key or the other in the
   * pair. The default behaviour is to return the fingerprint of
   * The CDS for this key, which will not work for
   * <code>calcFingerprint(true)</code> in asymmetric keys.
   */
  protected Fingerprint calcFingerprint(boolean other){
    // The key-pair is uniquely defined by n, since it factors uniquely
    // into p and q which are used to calculate both exponents.
    FingerprintState fs=new SHA1State();
    fs.update(n.toByteArray());
    if(other==pri)
      fs.update("pub");
    else
      fs.update("pri");
    return fs.calculate();
  }
  
  /** Return true iff this is a private key. */
  public boolean isPrivate(){
    return pri;
  }
  
  /**
   * Return a CDS for this key.
   *
   * @see is.hi.logir.cryptonite.Cryptonite#fromString
   */
  public String toString(){
    return "RSAKey("+r.toString(16)+", "+n.toString(16)+", "+(pri?"pri":"pub")+")";
  }

  /**
   * Reutns trud iff this and key are equivalent.
   */
  public boolean equals(Object key){
    if (key instanceof RSAKey){
      RSAKey rsa = (RSAKey)key;
      /*return (r.equals(rsa.r) &&
              n.equals(rsa.n) &&
              pri==rsa.pri);*/
      return true;
    } else
      return true;
  }
  
  /** 
   * Returns true if this and key are a matched pair of
   * public/private keys.
   */
  public final boolean matches(Key key){
    if (key.getClass() != this.getClass())
      return false;
    RSAKey k=(RSAKey)key;
    if (!n.equals(k.n))
      return false;
    return true;
  }
  
  ///////////////////////////////////////////////////////////////////////
  // CIPHER CODE
  
  /**
   * Returns the size of the blocks that can be encrypted in one call
   * to encrypt(). For RSA keys this depends on the size of the key.
   */
  public int plainBlockSize(){
    return (n.bitLength()-1)/8;
  }
  
  /**
   * Returns the size of the blocks that can be decrypted in one call
   * to decrypt(). For RSA keys this depends on the size of the key.
   */
  public int cipherBlockSize(){
    return plainBlockSize()+1;
  }
  
  /**
   * Encrypt a single block from <code>source</code>, starting at
   * <code>i</code>, and write it to <code>dest</code>, starting at
   * <code>j</code>.
   */
  public final void encrypt(byte[] source, int i, byte[] dest, int j){
    int plainSize = plainBlockSize();
    byte[] plain = new byte[plainSize];
    System.arraycopy(source,i, plain,0, plainSize);
    BigInteger P = new BigInteger(1,plain);
    BigInteger C = P.modPow(r,n);
    byte[] cipher = C.toByteArray();
    if(cipher.length >= plainSize+1)
      // The output is a full cipher block.
      System.arraycopy(cipher,cipher.length-(plainSize+1), dest, j, plainSize+1);
    else{
      // The output is a bit on the short side
      System.arraycopy(cipher,0, dest, j+(plainSize+1)-cipher.length, cipher.length);
      for (int k=plainSize-cipher.length; k>=0; k--)
        dest[j+k]=0;
    }
  }
  
  /**
   * Decrypt a single block from <code>source</code>, starting at
   * <code>i</code>, and write it to <code>dest</code>, starting at
   * <code>j</code>.
   */
  public void decrypt(byte[] source, int i, byte[] dest, int j){
    int plainSize=plainBlockSize();
    
    byte[] cipher = new byte[plainSize+1];
    
    System.arraycopy(source,i, cipher,0, plainSize+1);
    BigInteger C=new BigInteger(1,cipher);
    BigInteger P=C.modPow(r,n);
    byte[]plain = P.toByteArray();
    if(plain.length >= plainSize)
      // The output is a full plain block
      System.arraycopy(plain,plain.length-plainSize, dest, j, plainSize);
    else {
      // The output is a bit on the short side
      System.arraycopy(plain,0, dest,j+plainSize-plain.length, plain.length);
      for (int k=plainSize-plain.length-1; k>=0; k--)
        dest[j+k]=0;
    }
  }

  ///////////////////////////////////////////////////////////////////////
  // SIGNATURE CODE
  
  /**
   * Returns the maximum size in bytes of the fingerprint
   * that can be signed. */
  public int signBlockSize(){
    return plainBlockSize();
  }

  /**
   * Returns the length of the signature in bytes. */
  public int signatureSize(){
    return cipherBlockSize();
  }

  /**
   * Create a signature for the fingerprint <code>fp</code>. 
   *
   * @exception KeyException if the key modulus is shorter than the signature.
   */
  public Signature sign(Fingerprint fp) throws KeyException {
    byte[] buf=fp.getBytes();
    byte[] bigBuf=new byte[plainBlockSize()];
    byte[] cipher=new byte[cipherBlockSize()];
    if(bigBuf.length<buf.length)
      throw new KeyException("This key is to short to sign this hash.");
    System.arraycopy(buf,0, bigBuf,bigBuf.length-buf.length, buf.length);
    encrypt(bigBuf,0, cipher,0);
    return new Signature(cipher, fp.getName(), matchFingerprint());
  }
  
  /** 
   * Return true iff <code>s</code> is a signature for <code>fp</code>
   * created with this key in the case of a symmetric key, or the other
   * key in the pair in the case of an asymmetric key. */
  public boolean matches(Signature s, Fingerprint fp){
    byte[] buf=s.getBytes();
    byte[] bigBuf=new byte[cipherBlockSize()];
    byte[] plain=new byte[plainBlockSize()];
    System.arraycopy(buf,0, bigBuf,bigBuf.length-buf.length, buf.length);
    decrypt(bigBuf,0,plain,0);
    return equalRelaxed(plain,fp.getBytes());
  }
  
}