// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

import java.io.*;

/**
 * This OutputStream signs everything written to it using the specified
 * FingerprintState and SignatureKey.
 *
 * @see is.hi.logir.cryptonite.VerifyStream
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class SignStream extends FilterOutputStream{

  private FingerprintState fs;
  private SignatureKey key;

  private int bufPos;   // number of bytes written to current block
  private int bufSize;  // number of bytes in the block

  /**
   * Creates a new SignStream with out as its underlying stream. It passes
   * everything written to it through <code>fs</code> and after each
   * approximately <code>blockSize</code> bytes it inserts a signature of the
   * fingerprint into the underlying stream.
   * <p>
   * If <code>key</code> is null the fingerprints will be written unsigned
   * to the underlying stream.
   */
  public SignStream(OutputStream out, int blockSize, SignatureKey key, FingerprintState fs) {
    super(out);
    this.key = key;
    this.bufSize=(blockSize/fs.blockSize())*fs.blockSize();
    if (this.bufSize==0)
      this.bufSize=fs.blockSize();
    this.fs = fs;
  }

  /** Writes the specified byte to this output stream. */
  public synchronized void write(int b) throws IOException{
    byte[] buf = { (byte)b };
    write(buf,0,1);
  }

  /**
   * Writes <code>len</code> bytes from the specified byte array starting
   * at offset <code>off</code> to this output stream.
   *
   * @exception IOException if there is a problem iwth the underlying stream
   * or the key fails to sign the fingerprint. */
  public synchronized void write(byte[] buf, int off, int len) throws IOException{
    // REKUZUN RU13Z              (sigh... don't code before breakfast)
    if(len<=bufSize-bufPos){
      // It's <= one block
      out.write(buf,off,len);
      fs.update(buf,off,len);
      bufPos+=len;
      if(bufPos==bufSize){
        Fingerprint fp=fs.calculate();
        if(key==null)
          out.write(fp.getBytes());
        else
          try {
            Signature sig=key.sign(fp);
            out.write(sig.getBytes());
          } catch (KeyException e) {
            throw new IOException("KeyException: "+e.getMessage());
          }
        bufPos=0;
      }
    } else {
      // It's more than one block
      int l=bufSize-bufPos;
      write(buf,off,l);
      write(buf,off+l,len-l);
    }
  }
  
  /**
   * Flushes this output stream and forces any buffered output bytes to
   * be written out to the stream. If the number of bytes written is not
   * a multiple of the plainBlockSize of the Key used for encryption, up
   * to one byte less than a whole block of garbage may be appended to
   * the data when flush is called.
   */
  public synchronized void flush() throws IOException{
    out.flush();
  }

  /**
   * @exception IOException if there is a problem iwth the underlying stream
   * or the key fails to sign the fingerprint. */
  public synchronized void close() throws IOException{
    Fingerprint fp=fs.calculate();
    if(key==null)
      out.write(fp.getBytes());
    else
      try {
        Signature sig=key.sign(fp);
        out.write(sig.getBytes());
      } catch (KeyException e) {
        throw new IOException("KeyException: "+e.getMessage());
      }
    out.close();
  }

}
