// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite;

/**
 * This is the class for Triple-DES keys.
 * <p>
 * The CDS for a triple-DES key is <code>TriDESKey(key)</code> to create a specific key
 * or <code>TriDESKey(?)</code> for a random TriDESKey object.
 */
public class TriDESKey extends SymmetricKey implements CipherKey {

  private DESKey k1;
  private DESKey k2;
  private DESKey k3;

  /** Create a new random triple-DES key. */
  public TriDESKey(){
    k1 = new DESKey();
    k2 = new DESKey();
    k3 = new DESKey();
  }
  
  /** Create a new triple-DES key with the key bits from <code>key[0..23]</code>. */
  public TriDESKey(byte[] key){
    k1=new DESKey(makeLong(key, 0,8));
    k2=new DESKey(makeLong(key, 8,8));
    k3=new DESKey(makeLong(key,16,8));
  }
  
  /**
   * Called by <code>Cryptonite.fromString(String)</code> .
   *
   * @see is.hi.logir.cryptonite.Cryptonite#fromString
   *
   * @exception InvalidCDSException if the CDS is in some way malformed.
   */
  public TriDESKey(String key) throws InvalidCDSException{
    int i=0;
    while((i<key.length()) && Character.isWhitespace(key.charAt(i)))
      i++;
    if(i==key.length())
      throw new InvalidCDSException("Empty argument passed to TriDESKey(String)");
    int j=key.length()-1;
    while((j>=0) && Character.isWhitespace(key.charAt(j)))
      j--;
    // key[i..j] is the contents of key[] with spaces stripped.
    if(j==i && key.charAt(i)=='?'){
      k1 = new DESKey();
      k2 = new DESKey();
      k3 = new DESKey();
    }
    else {
      byte[] k = fromHexString(key);
      k1=new DESKey(makeLong(k, 0,8));
      k2=new DESKey(makeLong(k, 8,8));
      k3=new DESKey(makeLong(k,16,8));
    }
  }
  
  /** Return 8. */
  public int plainBlockSize(){
    return 8;
  }
  
  /** Return 8. */
  public int cipherBlockSize(){
    return 8;
  }
  
  /** Return 168 */
  public int getSize(){
    return 168;
  }
  
  /** Return "TriDES". */
  public String getAlgorithm(){
    return "TriDES";
  }
  
  /** Return true iff this is equivalent to <code>k</code>. */
  public boolean equals(Object o){
    if (o.getClass() != this.getClass())
      return false;
    TriDESKey k=(TriDESKey)o;
    return k1.equals(k.k1) && k1.equals(k.k2) && k1.equals(k.k3);
  }
  
  /** Return the key-bits for this key as an array of 24 bytes. */
  public byte[] getKey(){
    byte[] k=new byte[24];
    System.arraycopy(k1.getKey(),0, k, 0, 8);
    System.arraycopy(k2.getKey(),0, k, 8, 8);
    System.arraycopy(k3.getKey(),0, k,16, 8);
    return k;
  }
  
  /**
   * Return a CDS for this key.
   *
   * @see is.hi.logir.cryptonite.Cryptonite#fromString
   */
  public String toString(){
    StringBuffer sb=new StringBuffer(56);
    sb.append("TriDESKey(");
    sb.append(hexString(k1.getKey()));
    sb.append(hexString(k2.getKey()));
    sb.append(hexString(k3.getKey()));
    sb.append(")");
    return sb.toString();
  }
  
  /**
   * Encrypt one block of data. The plain data is taken from
   * <code>source[i..i+7]</code> and ciphertext is written to
   * <code>dest[j..j+7]</code>.
   */
  public void encrypt(byte[] source, int i, byte[] dest, int j){
    long block = makeLong(source,i,8);
    block = pickBits(block,DESKey.IP);
    block = k1.subCrypt(block);
    block = k2.subDecrypt(block);
    block = k3.subCrypt(block);
    block = pickBits(block,DESKey.FP);
    writeBytes(block,dest,j,8);
  }
  
  /**
   * Decrypt one block of data. The encrypted data is taken from
   * <code>dest[i..i+7]</code> and plaintext is written to
   * <code>source[j..j+7]</code>.
   */
  public void decrypt(byte[] source, int i, byte[] dest, int j){
    long block = makeLong(source,i,8);
    block = pickBits(block,DESKey.IP);
    block = k3.subDecrypt(block);
    block = k2.subCrypt(block);
    block = k1.subDecrypt(block);
    block = pickBits(block,DESKey.FP);
    writeBytes(block, dest, j, 8);
  }

}
