// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite.test;

import is.hi.logir.cryptonite.*;
import java.util.Random;

/**
 * This application tests the various Key classes.
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 *
 * @see is.hi.logir.cryptonite.Key
 */
public class TestKey extends Cryptonite{

  // Insecure but fast pseud-random number generator
  static Random rand=new Random();  
  
  public static void help(){
    System.out.println("Use: java is.hi.logir.cryptonite.test.TestKey <type>");
    System.out.println("where <type> is one of DES, TriDES, RSA or Caesar.");
  }

  public static KeyPair createKeys(String[] arg){
    System.out.println("\nKEY GENERATION");
    if(arg.length==0){
      help();
    } else if(arg[0].equals("DES")){
      Key k = new DESKey();
      return new KeyPair(k,k);
    } else if(arg[0].equals("TriDES")){
      Key k = new TriDESKey();
      return new KeyPair(k,k);
    } else if(arg[0].equals("RSA")){
      return RSAKey.createKeys(512);
    } else if(arg[0].equals("Caesar")){
      Key k = new CaesarKey();
      return new KeyPair(k,k);
    } else
      help();
    return null;
  }
  
  public static void printKeys(Key pub, Key pri){
    if(pri instanceof SymmetricKey){
      System.out.println(pub);
    } else {
      System.out.println(pub);
      System.out.println(pri);
    }
  }

  public static boolean testCDS(Key pub, Key pri){
    System.out.println("\nTESTING CONVERSION TO AND FROM STRINGS");
    boolean ok;
    try{
      String cds=pub.toString();
      Key k = (Key)fromString(cds);
      System.out.println(k);
      ok = pub.equals(k);
      if(!(pri instanceof SymmetricKey)){
        cds=pri.toString();
        k = (Key)fromString(cds);
        System.out.println(k);
        ok = ok && pri.equals(k);
      }
    } catch (Throwable e){
      System.out.println(e);
      e.printStackTrace(System.out);
      ok = false;
    }
    System.out.println(ok ? "Passed" : "Failed" );
    return ok;
  }

  public static boolean testCrypt(CipherKey pub, CipherKey pri){
    System.out.println("\nTESTING ENCRYPTION/DECRYPTION");
    boolean allOk=true;
    try{
      byte[] plain  = new byte[pub.plainBlockSize()];
      byte[] plain2 = new byte[pub.plainBlockSize()];
      byte[] cipher = new byte[pub.cipherBlockSize()];
      System.out.println("Plain/cipher block sizes = "
                         +pub.plainBlockSize()+", "
                         +pub.cipherBlockSize());
      
      System.out.println("Encrypting/decrypting 4KB of random data:");
      int max = 4096/pub.plainBlockSize();
      java.util.Random rand = new java.util.Random();
      int failCount=0;
      for (int i=0; i<max; i++){
        rand.nextBytes(plain);
        rand.nextBytes(plain2);
        rand.nextBytes(cipher);
        pub.encrypt(plain,0, cipher,0);
        pri.decrypt(cipher,0, plain2,0);
        boolean ok = (plain.length == plain2.length);
        for(int j=0; j<plain.length; j++)
          if (plain[j]!=plain2[j]){
            ok=false;
            failCount++;
            break;
          }
        System.out.print(ok ? "+" : "-");
        allOk=(allOk && ok);
      }
      System.out.println();
    } catch (Throwable e){
      System.out.println(e);
      e.printStackTrace(System.out);
      allOk = false;
    }
    System.out.println("Test "+ (allOk?"passed":"failed"));
    return allOk;
  }

  public static boolean testSign(SignatureKey pub, SignatureKey pri){
    System.out.println();
    boolean ok;
    try {
      byte[] buffer = new byte[1024];
      rand.nextBytes(buffer);
      Fingerprint fp=Fingerprint.create(buffer,"SHA1");
      System.out.println("Good signature:");
      Signature s1=pri.sign(fp);
      System.out.println(s1);
      boolean m1 = pub.matches(s1,fp);
      System.out.println("Signature verified: "+(m1?"Yes":"No"));

      System.out.println();
      System.out.println("Bogus signature:");
      byte[] bogus = s1.getBytes();
      bogus[3]^=5;
      Signature s2=new Signature(bogus, "SHA1", pub.getFingerprint());
      System.out.println(s2);
      boolean m2=pub.matches(s2,fp);
      System.out.println("Signature verified: "+(m2?"Yes":"No"));

      ok = m1 && !m2;
    } catch (Throwable e){
      System.out.println(e);
      e.printStackTrace(System.out);
      ok = false;
    }
    System.out.println(ok ? "Passed" : "Failed" );
    return ok;
  }

  /** Run with argument DES, TriDES, RSA or Caesar. */
  public static void main(String[] arg){
    KeyPair kp = createKeys(arg);
    if (kp==null)
      return;
    Key pub = kp.getPublic();
    Key pri = kp.getPrivate();
    printKeys(pub,pri);

    boolean ok=true;
    
    ok &= testCDS(pub,pri);

    if(pub instanceof CipherKey)
      ok &= testCrypt((CipherKey)pub,(CipherKey)pri);

    if(pub instanceof SignatureKey)
      ok &= testSign((SignatureKey)pub,(SignatureKey)pri);

    System.out.println("\nALL TESTS");
    System.out.println(ok ? "Passed" : "Failed" );
  }
  
}