// Copyright (C) 1998 Logi Ragnarsson

package is.hi.logir.cryptonite.test;

import is.hi.logir.cryptonite.*;
import java.util.Random;

/**
 * This application tests the various encryption-mode classes.
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 *
 * @see is.hi.logir.cryptonite.EncryptMode
 */
public class TestMode extends Cryptonite{
  
  // Insecure but fast pseud-random number generator
  static Random rand=new Random();  
  

  // Tests performed. + passed, - failed.
  
  // M/C     DES  TriDES RSA
  // ECB      +     +     +
  // CBC      +     +     +
  // CFB      +     +     +
  // OFB      +     +     +
  
  public static void help(){
    System.out.println("Use: java is.hi.logir.cryptonite.test.TestKey <key-type> <mode>");
    System.out.println("where <key-type> is one of DES, TriDES, RSA or Caesar.");
    System.out.println("and   <mode> is one of ECB, CBC, CFB or OFB.");
  }
  
  public static KeyPair createKeys(String[] arg){
    System.out.println("\nKEY GENERATION");
    if(arg.length==0){
      help();
    } else if(arg[0].equals("DES")){
      Key k = new DESKey();
      return new KeyPair(k,k);
    } else if(arg[0].equals("TriDES")){
      Key k = new TriDESKey();
      return new KeyPair(k,k);
    } else if(arg[0].equals("RSA")){
      return RSAKey.createKeys(256);
    } else if(arg[0].equals("Caesar")){
      Key k = new CaesarKey();
      return new KeyPair(k,k);
    } else
      help();
    return null;
  }
  
  public static EncryptMode createEncrypt(String[] arg, CipherKey pub, CipherKey pri){
    if(arg.length<2){
      help();
    } else if (arg[1].equals("ECB")){
      return new EncryptECB(pub);
    } else if (arg[1].equals("CBC")){
      return new EncryptCBC(pub);
    } else if (arg[1].equals("CFB")){
      return new EncryptCFB(pri);
    } else if (arg[1].equals("OFB")){
      return new EncryptOFB(pri,256);
    } else
      help();
    
    return null;
  }
  
  public static DecryptMode createDecrypt(String[] arg, CipherKey pub, CipherKey pri){
    if(arg.length<2){
      help();
    } else if (arg[1].equals("ECB")){
      return new DecryptECB(pri);
    } else if (arg[1].equals("CBC")){
      return new DecryptCBC(pri);
    } else if (arg[1].equals("CFB")){
      return new DecryptCFB(pri);
    } else if (arg[1].equals("OFB")){
      return new DecryptOFB(pri,256);
    } else
      help();
    
    return null;
  }
  
  public static boolean test(EncryptMode em, DecryptMode dm){
    System.out.println("\nTESTING MODE");
    Random rand = new Random();
    
    int SIZE=8192;
    byte[] source = new byte[SIZE];
    byte[] dest   = new byte[SIZE];
    int i=0;
    int j=0;
    int k=0;
    rand.nextBytes(source);
    
    System.out.println("Encrypting "+SIZE/1024+" KB.");
    while(j<SIZE){
      byte[] t1;
      if (i<SIZE){
        // there is data left to encrypt
        int n = Math.min(rand.nextInt()%64+64, SIZE-i);
        t1 = em.encrypt(source,i,n);
        i+=n;
      } else
        t1 = em.flush();

      byte[] t2 = dm.decrypt(t1,0,t1.length);
      int n=Math.min(t2.length,SIZE-j);
      System.arraycopy(t2,0, dest,j, n);
      j+=n;
      
      while(k<j){
        if (dest[k]!=source[k]){
          System.err.println("Error at offset "+k);
          return false;
        }
        k++;
      }
      System.out.print(".");
      System.out.flush();
    }
    return true;
  }
  
  
  /** Run with arguments DES, TriDES, RSA or Caesar and ECB, CBC, CFB or OFB. */
  public static void main(String[] arg){
    KeyPair kp = createKeys(arg);
    if (kp==null)
      return;

    // For asymmetric ciphers, pub will be a public key. Other wise pub==pri.
    Key pub = kp.getPublic();
    Key pri = kp.getPrivate();

    EncryptMode em = createEncrypt(arg,(CipherKey)pub,(CipherKey)pri);
    if (em==null)
      return;
    DecryptMode dm = createDecrypt(arg,(CipherKey)pub,(CipherKey)pri);
    if (em==null)
      return;
    
    boolean ok = test(em,dm);
    
    System.out.println();
    System.out.println(ok ? "Passed" : "Failed" );
  }
  
}